/****************************************************************************/
/* This module is based on xeyes, but has been siginificantly modified      */
/* by INOUE Seiichiro (inoue@ainet.or.jp)			                        */
/****************************************************************************/
/* $XConsortium: Eyes.c,v 1.26 91/08/23 12:26:40 gildea Exp $ */
/*
 * Copyright 1991 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/*
 * Eyes.c
 *
 * a widget which follows the mouse around
 */

#include <X11/Xos.h>
#include <stdio.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Xmu/Converters.h>
#include <X11/extensions/shape.h>
#include <X11/xpm.h>
#include "EyesP.h"
#include "xeyes.h"
#include "readdata.h"
#include <math.h>

#if (defined(SVR4) || defined(SYSV) && defined(SYSV386)) && __STDC__
extern double hypot(double, double);
#endif

#define offset(field) XtOffsetOf(EyesRec, eyes.field)
#define goffset(field) XtOffsetOf(WidgetRec, core.field)

static XtResource resources[] = {
    {XtNwidth, XtCWidth, XtRDimension, sizeof(Dimension),
	goffset(width), XtRImmediate, (XtPointer) 150},
    {XtNheight, XtCHeight, XtRDimension, sizeof(Dimension),
	goffset(height), XtRImmediate, (XtPointer) 100},
    {XtNbackingStore, XtCBackingStore, XtRBackingStore, sizeof (int),
    	offset (backing_store), XtRString, "default"},
};

#undef offset
#undef goffset


#define TPointEqual(a, b)  ((a).x == (b).x && (a).y == (b).y)
#define XPointEqual(a, b)  ((a).x == (b).x && (a).y == (b).y)

static int delays[] = { 50, 100, 200, 400, 0 };

/* Private function prototype declarations */
static void ClassInitialize();
static void Initialize(Widget greq, Widget gnew, ArgList args, Cardinal *num_args);
static void Resize(Widget gw);
static void Realize(Widget gw, XtValueMask *valueMask, XSetWindowAttributes *attrs);
static void Destroy(Widget gw);
static void Redisplay(Widget gw, XEvent event, Region region);
static Boolean SetValues(Widget current, Widget request, Widget new,
						 ArgList args, Cardinal *num_args);

static TPoint computeNewPosition(const TAnimArea *taArea, TPoint mouse);
static void draw_it(XtPointer client_data, XtIntervalId *id);
static void animation(EyesWidget w, int num);



EyesClassRec eyesClassRec = {
    { /* core fields */
    /* superclass		*/	&widgetClassRec,
    /* class_name		*/	"Eyes",
    /* size			*/	sizeof(EyesRec),
    /* class_initialize		*/	ClassInitialize,
    /* class_part_initialize	*/	NULL,
    /* class_inited		*/	FALSE,
    /* initialize		*/	Initialize,
    /* initialize_hook		*/	NULL,
    /* realize			*/	Realize,
    /* actions			*/	NULL,
    /* num_actions		*/	0,
    /* resources		*/	resources,
    /* num_resources		*/	XtNumber(resources),
    /* xrm_class		*/	NULLQUARK,
    /* compress_motion		*/	TRUE,
    /* compress_exposure	*/	TRUE,
    /* compress_enterleave	*/	TRUE,
    /* visible_interest		*/	FALSE,
    /* destroy			*/	Destroy,
    /* resize			*/	Resize,
    /* expose			*/	Redisplay,
    /* set_values		*/	SetValues,
    /* set_values_hook		*/	NULL,
    /* set_values_almost	*/	NULL,
    /* get_values_hook		*/	NULL,
    /* accept_focus		*/	NULL,
    /* version			*/	XtVersion,
    /* callback_private		*/	NULL,
    /* tm_table			*/	NULL,
    /* query_geometry		*/	XtInheritQueryGeometry,
    }
};

static void ClassInitialize()
{
    XtAddConverter(XtRString, XtRBackingStore, XmuCvtStringToBackingStore,
				   NULL, 0);
}

WidgetClass eyesWidgetClass = (WidgetClass) &eyesClassRec;


static void
Initialize(Widget greq, Widget gnew, ArgList args, Cardinal *num_args)
{
    EyesWidget w = (EyesWidget)gnew;
    XtGCMask	valuemask;
    XGCValues	myXGCV;
    int 		shape_event_base, shape_error_base;
	int			num;
	int			aWidth, aHeight;
	int			cX, cY;
	char		data[VALUE_MAX];
	int			memlen;
	int			ret;
	

	if (getvalue(data, ANIMNUM) == False) {
		/* No anim data */
		fprintf(stderr, "%s : No data \n", ANIMNUM);
		exit(1);
	}
	sscanf(data, "%d", &w->eyes.num_anim);
	memlen = sizeof(AnimInfo) * w->eyes.num_anim;
	if ((w->eyes.anim = (AnimInfo*)malloc(memlen)) == NULL) {
		perror("malloc");
		exit(1);
	}
	memset(w->eyes.anim, 0, memlen);
	
	for (num = 0; num < w->eyes.num_anim; num++) {
		if (getvalue(data, ANIMPIXMAP) == False) {
			/* No more anim data */
			w->eyes.num_anim = num;
			break;
		}
		ret = XpmReadFileToPixmap(dpy, rootwindow,
								  EnrootPathName(data),
								  &w->eyes.anim[num].pix,
								  &w->eyes.anim[num].mask,
								  &w->eyes.anim[num].att);
		if (ret != XpmSuccess) {
			xpm_error_handler(ret, data);
		}

		if (getvalue(data, ANIMAREA) == True)
			sscanf(data, "%d, %d", &aWidth, &aHeight);
		else {
			/* default value */
			aWidth = w->core.width;
			aHeight = w->core.height;
		}
		if (getvalue(data, ANIMCENTER) == True)
			sscanf(data, "%d, %d", &cX, &cY);
		else {
			/* default value */
			cX = w->core.width / 2 ;
			cY = w->core.height / 2;
		}
		SetTransform(&w->eyes.anim[num].t,
					 0, aWidth,
					 aHeight, 0,
					 -1.0, 1.0,
					 -1.0, 1.0);
		w->eyes.anim[num].animflag =
			SetTAnimArea(&w->eyes.anim[num].taArea, 
						 aWidth, aHeight,
						 w->eyes.anim[num].att.width, w->eyes.anim[num].att.height,
						 cX, cY);
		w->eyes.anim[num].pos.x = -1000;
		w->eyes.anim[num].pos.y = -1000;
	}
	dataterm();	

    w->eyes.update = 0;
    /* wait for Realize to add the timeout */
    w->eyes.interval_id = 0;

	return;
}

static void
Resize(Widget gw)
{
    EyesWidget	w = (EyesWidget)gw;

    if (XtIsRealized(gw)) {
    	XClearWindow(dpy, XtWindow(w));
    }
}

static void
Realize(Widget gw, XtValueMask *valueMask, XSetWindowAttributes *attrs)
{
    EyesWidget	w = (EyesWidget)gw;
	Window	win;
	int		num;
	
    if (w->eyes.backing_store != Always + WhenMapped + NotUseful) {
     	attrs->backing_store = w->eyes.backing_store;
		*valueMask |= CWBackingStore;
    }
    XtCreateWindow(gw, (unsigned)InputOutput, (Visual *)CopyFromParent,
				   *valueMask, attrs);

	win = XtWindow(gw);
	/* Create GCs */
	for (num = 0; num < w->eyes.num_anim; num++) {
		w->eyes.anim[num].gc = XCreateGC(dpy, win, 0, NULL);
		XSetClipMask(dpy, w->eyes.anim[num].gc, w->eyes.anim[num].mask);
	}
	w->eyes.baseGC = XCreateGC(dpy, win, 0, NULL);

	/* Set timer */
	if (w->eyes.num_anim > 0) {
		w->eyes.interval_id =
			XtAppAddTimeOut(XtWidgetToApplicationContext(gw),
							delays[w->eyes.update], draw_it, (XtPointer)gw);
	}

	return;
}

static void
Destroy(Widget gw)
{
     EyesWidget w = (EyesWidget)gw;
	 int	num;

     if (w->eyes.interval_id)
		 XtRemoveTimeOut (w->eyes.interval_id);
	 for (num = 0; num < w->eyes.num_anim; num++) {
		 XFreePixmap(dpy, w->eyes.anim[num].pix);
		 XFreePixmap(dpy, w->eyes.anim[num].mask);
		 XFreeGC(dpy, w->eyes.anim[num].gc);
	 }
     XFreeGC(dpy, w->eyes.baseGC);
}

/* ARGSUSED */
static void
Redisplay(Widget gw, XEvent event, Region region)
{
    EyesWidget	w = (EyesWidget)gw;
	Window	win = XtWindow(gw);
	int		num;
	
	if (XtIsRealized(gw)) {
		XCopyArea(dpy, base_pix, win, w->eyes.baseGC,
				  0, 0, w->core.width, w->core.height, 0, 0);
		for (num = 0; num < w->eyes.num_anim; num++) {
			animation(w, num);
		}
	}
}

/* ARGSUSED */
static Boolean
SetValues(Widget current, Widget request, Widget new,
		  ArgList args, Cardinal *num_args)
{
    return FALSE;
}


/*
  static TPoint computeNewPosition(const TAnimArea *taArea, TPoint mouse)

  INPUT:
  const TAnimArea *taArea;
  TPoint mouse;
  OUTPUT:
  Return value;	computed new position
  */
static TPoint
computeNewPosition(const TAnimArea *taArea, TPoint mouse)
{
	double	cx, cy;
	double	dist;
	double	angle;
	double	x, y;
	double	h;
	double	dx, dy;
	double	cosa, sina;
	TPoint	ret;

	dx = mouse.x - taArea->cx;
	dy = mouse.y - taArea->cy;
	if (dx == 0 && dy == 0) {
		cx = taArea->cx;
		cy = taArea->cy;
	} else {
		angle = atan2((double)dy, (double)dx);
		cosa = cos(angle);
		sina = sin(angle);
		h = hypot(taArea->hh * cosa, taArea->hw * sina);
		x = (taArea->hw * taArea->hh) * cosa / h;
		y = (taArea->hw * taArea->hh) * sina / h;
		dist = taArea->dist * hypot(x, y);
		if (dist > hypot((double)dx, (double)dy)) {
			cx = dx + taArea->cx;
			cy = dy + taArea->cy;
		} else {
			cx = dist * cosa + taArea->cx;
			cy = dist * sina + taArea->cy;
		}
	}
	ret.x = cx;
	ret.y = cy;
	return ret;
}


/* ARGSUSED */
static void
draw_it(XtPointer client_data, XtIntervalId *id)
{
	EyesWidget	w = (EyesWidget)client_data;
	Window		rep_root, rep_child;
	int		rep_rootx, rep_rooty;
	unsigned int	rep_mask;
	int		dx, dy;
	TPoint		mouse;
	Window		win = XtWindow (w);
	TPoint		newpos;
	XPoint		xnewpos;
	XPoint		xpos;
	int			num;

	if (XtIsRealized((Widget)w)) {
		XQueryPointer(dpy, win, &rep_root, &rep_child,
					  &rep_rootx, &rep_rooty, &dx, &dy, &rep_mask);
		mouse.x = Tx(dx, dy, &w->eyes.anim[0].t);
		mouse.y = Ty(dx, dy, &w->eyes.anim[0].t);
		if (!TPointEqual(mouse, w->eyes.mouse)) {
			w->eyes.mouse = mouse;	/* keep */
			for (num = 0; num < w->eyes.num_anim; num++) {
				if (w->eyes.anim[num].animflag == False)
					continue;
				mouse.x = Tx(dx, dy, &w->eyes.anim[num].t);
				mouse.y = Ty(dx, dy, &w->eyes.anim[num].t);
				newpos = computeNewPosition(&w->eyes.anim[num].taArea, mouse);
				xpos.x = Xx(w->eyes.anim[num].pos.x, w->eyes.anim[num].pos.y,
							&w->eyes.anim[num].t);
				xpos.y = Xy(w->eyes.anim[num].pos.x, w->eyes.anim[num].pos.y,
							&w->eyes.anim[num].t);
				xnewpos.x = Xx(newpos.x, newpos.y,
							   &w->eyes.anim[num].t);
				xnewpos.y = Xy(newpos.x, newpos.y,
							   &w->eyes.anim[num].t);
				if (!XPointEqual(xpos, xnewpos)) {
					newpos.x -= w->eyes.anim[num].taArea.offw;
					newpos.y -= w->eyes.anim[num].taArea.offh;
					if (w->eyes.anim[num].pos.x == -1000)
						w->eyes.anim[num].pos = newpos;
					w->eyes.anim[num].old_pos = w->eyes.anim[num].pos;
					w->eyes.anim[num].pos = newpos;
					animation(w, num);
				}
			}
			w->eyes.update = 0;
		}			
		else {
			if (delays[w->eyes.update + 1] != 0)
				++w->eyes.update;
		}
	}
	w->eyes.interval_id =
		XtAppAddTimeOut(XtWidgetToApplicationContext((Widget) w),
				delays[w->eyes.update], draw_it, (XtPointer)w);
} /* draw_it */
    

static void animation(EyesWidget w, int num)
{
	Window win = XtWindow(w);

	/* base pixmap */
	TCopyArea(base_pix, win, w->eyes.baseGC,
			  &w->eyes.anim[num].t,
			  w->eyes.anim[num].old_pos.x, w->eyes.anim[num].old_pos.y,
			  w->eyes.anim[num].att.width, w->eyes.anim[num].att.height,
			  w->eyes.anim[num].old_pos.x, w->eyes.anim[num].old_pos.y);
	/* animation pixmap */
	TSetClipOrigin(w->eyes.anim[num].gc,
				   w->eyes.anim[num].pos.x, w->eyes.anim[num].pos.y,
				   &w->eyes.anim[num].t);
	TCopyArea2(w->eyes.anim[num].pix, win, w->eyes.anim[num].gc,
			   &w->eyes.anim[num].t,
			   0, 0,
			   w->eyes.anim[num].att.width, w->eyes.anim[num].att.height,
			   w->eyes.anim[num].pos.x, w->eyes.anim[num].pos.y);
}
