/****************************************************************************/
/* This module is based on xeyes, but has been siginificantly modified      */
/* by INOUE Seiichiro (inoue@ainet.or.jp)			                        */
/****************************************************************************/
/*
 * transformed coordinate system objects for X
 */

#include	<X11/Xlib.h>
#include	"xeyes.h"
#include	"transform.h"


static XPoint *
TranslatePoints(TPoint *points, int n_points, Transform *t, int mode)
{
	XPoint	*xpoints;
	int	i;
	double	xoff = 0.0, yoff = 0.0;

	xpoints = (XPoint *) malloc (n_points * sizeof (*xpoints));
	if (!xpoints)
		return 0;
	for (i = 0; i < n_points; i++) {
		xpoints[i].x = Xx(points[i].x + xoff, points[i].y + yoff, t);
		xpoints[i].y = Xy(points[i].x + xoff, points[i].y + yoff, t);
		if (mode == CoordModePrevious) {
			xoff += points[i].x;
			yoff += points[i].y;
		}
	}
	return xpoints;
}

#if 0
TFillPolygon (dpy, d, gc, t, points, n_points, shape, mode)
register Display	*dpy;
Drawable		d;
GC			gc;
Transform		*t;
TPoint			*points;
int			n_points;
int			shape;
int			mode;
{
	XPoint	*xpoints;

	xpoints = TranslatePoints (points, n_points, t, mode);
	if (xpoints) {
		XFillPolygon (dpy, d, gc, xpoints, n_points, shape,
				CoordModeOrigin);
		free (xpoints);
	}
}

TDrawArc (dpy, d, gc, t, x, y, width, height, angle1, angle2)
	register Display	*dpy;
	Drawable		d;
	GC			gc;
	Transform		*t;
	double			x, y, width, height;
	int			angle1, angle2;
{
	int	xx, xy, xw, xh;

	xx = Xx(x,y,t);
	xy = Xy(x,y,t);
	xw = Xwidth (width, height, t);
	xh = Xheight (width, height, t);
	if (xw < 0) {
		xx += xw;
		xw = -xw;
	}
	if (xh < 0) {
		xy += xh;
		xh = -xh;
	}
	XDrawArc (dpy, d, gc, xx, xy, xw, xh, angle1, angle2);
}

TFillArc (dpy, d, gc, t, x, y, width, height, angle1, angle2)
	register Display	*dpy;
	Drawable		d;
	GC			gc;
	Transform		*t;
	double			x, y, width, height;
	int			angle1, angle2;
{
	int	xx, xy, xw, xh;

	xx = Xx(x,y,t);
	xy = Xy(x,y,t);
	xw = Xwidth (width, height, t);
	xh = Xheight (width, height, t);
	if (xw < 0) {
		xx += xw;
		xw = -xw;
	}
	if (xh < 0) {
		xy += xh;
		xh = -xh;
	}
	XFillArc (dpy, d, gc, xx, xy, xw, xh, angle1, angle2);
}
#endif	/* #if 0 */


void
TCopyArea(Drawable src, Drawable dest, GC gc, Transform *t,
		  double sx, double sy, int width, int height, double dx, double dy)
{
	int	src_x, src_y;
	int	dest_x, dest_y;

	src_x = Xx(sx, sy, t);
	src_y = Xy(sx, sy, t);
	dest_x = Xx(dx, dy, t);
	dest_y = Xy(dx, dy, t);
	XCopyArea(dpy, src, dest, gc, src_x, src_y, width, height, dest_x, dest_y);
}

void
TCopyArea2(Drawable src, Drawable dest, GC gc, Transform *t,
		   int sx, int sy, int width, int height, double dx, double dy)
{
	int	dest_x, dest_y;

	dest_x = Xx(dx, dy, t);
	dest_y = Xy(dx, dy, t);
	XCopyArea(dpy, src, dest, gc, sx, sy, width, height, dest_x, dest_y);
}

void
TSetClipOrigin(GC gc, double x, double y, Transform *t)
{
	int	xx, xy;

	xx = Xx(x,y,t);
	xy = Xy(x,y,t);
	XSetClipOrigin(dpy, gc, xx, xy);
}


void
SetTransform(Transform *t, int xx1, int xx2, int xy1, int xy2,
			 double tx1, double tx2, double ty1, double ty2)
{
	t->mx = ((double) xx2 - xx1) / (tx2 - tx1);
	t->bx = ((double) xx1) - t->mx * tx1;
	t->my = ((double) xy2 - xy1) / (ty2 - ty1);
	t->by = ((double) xy1) - t->my * ty1;
}


int
SetTAnimArea(TAnimArea *taArea,
			 int wArea, int hArea, int w, int h, int cx, int cy)
{
	/* center point  */
	taArea->cx = (double)cx / (wArea / 2.0) - 1.0;
	taArea->cy = -((double)cy / (hArea / 2.0) - 1.0);
	/* offset */
	taArea->offw = (double)w / wArea;
	taArea->offh = -(double) h / hArea;
	/* radius */
	taArea->hw = 1.0 - taArea->offw;
	taArea->hh = 1.0 + taArea->offh;
	/* distance */
	taArea->dist = 1.0;

	if (taArea->hw >0 && taArea->hh > 0)
		return True;
	else
		return False;
}
