/****************************************************************************/
/* xeyes generic															*/
/* This module is based on xeyes, but has been siginificantly modified      */
/* by INOUE Seiichiro (inoue@ainet.or.jp)			                        */
/****************************************************************************/
/* $XConsortium: xeyes.c,v 1.15 91/07/18 16:46:26 rws Exp $ */
/*
 * Copyright 1991 Massachusetts Institute of Technology
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation, and that the name of M.I.T. not be used in advertising or
 * publicity pertaining to distribution of the software without specific,
 * written prior permission.  M.I.T. makes no representations about the
 * suitability of this software for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * M.I.T. DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO EVENT SHALL M.I.T.
 * BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
 * OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
 * CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xmu/Editres.h>
#include <X11/extensions/shape.h>
#include <X11/xpm.h>
#include "Eyes.h"
#include "xeyes.h"
#include "readdata.h"
#include <stdio.h> 


/* Global variables */
Display *dpy;
Window	rootwindow;
Pixmap	base_pix;


/* Private variables */
static Atom wm_delete_window;

/* Private function prototype declarations */
static void quit(Widget w, XEvent *event, String *params, Cardinal *num_params);
static void usage(char *progname);


/* Application specific resources */
typedef struct {
	char	*datafile;
} AppResourcesRec;

#define XtNdatafile		"datafile"
#define XtCDatafile		"Datafile"
#define offset(field) XtOffsetOf(AppResourcesRec, field)
static XtResource      appResourceList[] = {
	{XtNdatafile, XtCDatafile, XtRString, sizeof(String),
    offset(datafile), XtRString, "xeyes.data"},
};
#undef offset


/* Command line options table.  Only resources are entered here...there is a
   pass over the remaining options after XtParseCommand is let loose. */
static XrmOptionDescRec options[] = {
	{"-file", "*datafile", XrmoptionSepArg, NULL},
	{"-data", "*datafile", XrmoptionSepArg, NULL},
	{"-backing",	"*eyes.backingStore",	XrmoptionSepArg,	NULL},
};

static XtActionsRec actions[] = {
    {"quit",	quit}
};



void main(int argc, char **argv)
{
    XtAppContext app_context;
	AppResourcesRec appResources;
    Arg arg[2];
    Cardinal i;
	Window	topw;
	Widget	topwg;
	Widget	eyewg;
	Pixmap	base_mask;
	XpmAttributes	base_att;
	char	basefname[VALUE_MAX];
    int		shape_event_base, shape_error_base;
	int		ret;
	

    topwg = XtAppInitialize(&app_context, "XEyes", 
							options, XtNumber(options), &argc, argv,
							NULL, NULL, 0);
    if (argc > 2)
		usage(argv[0]);
	
/*
	XtAddEventHandler(topwg, (EventMask)0, True, _XEditResCheckMessages, NULL);
*/
	dpy = XtDisplay(topwg);
	rootwindow = DefaultRootWindow(dpy);

	/* Get application resources */
	XtGetApplicationResources(topwg, (XtPointer)&appResources,
							  appResourceList, XtNumber(appResourceList),
							  NULL, 0);
	
    XtAppAddActions(app_context, actions, XtNumber(actions));
    XtOverrideTranslations(topwg,
						   XtParseTranslationTable("<Message>WM_PROTOCOLS: quit()"));
    
	/* data file */
    if (datainit(appResources.datafile) == False) {
		if (argc == 2) {
			if (datainit(argv[1]) == False) {
				fprintf(stderr, "Can't open files: \"%s\" and \"%s\" \n",
						appResources.datafile, argv[1]);
				usage(argv[0]);
				exit(1);
			}
		}
		else {
			fprintf(stderr, "Can't open file: \"%s\" \n",
					appResources.datafile);
			usage(argv[0]);
			exit(1);
		}
	}
	
	if (getvalue(basefname, BASEPIXMAP) == False) {
		fprintf(stderr, "%s : No data \n", BASEPIXMAP);
		exit(1);
	}
	/* Base image */
	memset(&base_att, 0, sizeof(base_att));
	ret = XpmReadFileToPixmap(dpy, rootwindow,
							  EnrootPathName(basefname),
							  &base_pix, &base_mask, &base_att);
	if (ret != XpmSuccess) {
		xpm_error_handler(ret, basefname);
	}
	
	/* Hope not to be resized */
    i = 0;
    XtSetArg(arg[i], XtNminWidth, base_att.width);
	i++;
    XtSetArg(arg[i], XtNminHeight, base_att.height);
	i++;
/*	
    i = 0;
    XtSetArg (arg[i], XtNiconPixmap, 
	      XCreateBitmapFromData (XtDisplay(toplevel),
				     XtScreen(toplevel)->root,
				     (char *)eyes_bits, eyes_width, eyes_height));
    i++;
    XtSetArg (arg[i], XtNiconMask, 
	      XCreateBitmapFromData (XtDisplay(toplevel),
				     XtScreen(toplevel)->root,
				     (char *)eyesmask_bits,
				     eyesmask_width, eyesmask_height));
    i++;
*/
    XtSetValues(topwg, arg, i);
	
    eyewg = XtVaCreateManagedWidget("eyes", eyesWidgetClass, topwg,
									XtNwidth, base_att.width,
									XtNheight, base_att.height,
									NULL);
    XtRealizeWidget(topwg);
	topw = XtWindow(topwg);

    wm_delete_window = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    (void)XSetWMProtocols(dpy, topw, &wm_delete_window, 1);

	if (XShapeQueryExtension(dpy, &shape_event_base, &shape_error_base)) {
		XShapeCombineMask(dpy, XtWindow(topwg),
						  ShapeBounding, 0, 0, base_mask, ShapeSet);
	}
	
    XtAppMainLoop(app_context);
}


/*ARGSUSED*/
static void
quit(Widget w, XEvent *event, String *params, Cardinal *num_params)
{
	dataterm();
	
    if (event->type == ClientMessage && 
		event->xclient.data.l[0] != wm_delete_window) {
		XBell(XtDisplay(w), 0);
    } else {
		XtDestroyApplicationContext(XtWidgetToApplicationContext(w));
		exit(0);
    }
}

/* Exit with message describing command line format */
static void
usage(char *progname)
{
    fprintf(stderr,
			"usage: %s\n", progname);
    fprintf(stderr,
			"	[datafile]\n");
    fprintf(stderr,
			"	[-data|-file {datafile}]\n");
    fprintf(stderr,
			"	[-geometry [{width}][x{height}][{+-}{xoff}[{+-}{yoff}]]] [-display [{host}]:[{vs}]]\n");
    fprintf(stderr, "\n");
    fprintf(stderr,
			"	[-backing {backing-store}]\n");
    exit(1);
}

  
void xpm_error_handler(int xpm_err, const char *filename)
{
	if (xpm_err == XpmSuccess)
		return;
	
	switch (xpm_err) {
	case XpmColorError:
		fprintf(stderr, "Warning: XpmColorError  \"%s\"\n", filename);
		return;

	case XpmOpenFailed:
		fprintf(stderr, "Error: XpmOpenFailed  \"%s\"\n", filename);
	case XpmFileInvalid:
		fprintf(stderr, "Error: XpmFileInvalid  \"%s\"\n", filename);
	case XpmNoMemory:
		fprintf(stderr, "Error: XpmNoMemory\n");
	case XpmColorFailed:
		fprintf(stderr, "Error: XpmColorFailed  \"%s\"\n", filename);
		exit(1);

	default:
		fprintf(stderr, "Unexpected xpm error \"%s\"\n", filename);
		exit(1);
	}
}
